//**************************************************************************
// 対象マイコン Renesas Electronics RMC_RA4M1 Rev.1.0bボード(R7FA4M1AB3CFP9 100pin)
//             ※RMC_RA4M1 Rev.2.0ボードは、ポート配置が換わるためこのプログラムでは動作しません
// ﾌｧｲﾙ内容     Basicマイコンカー走行プログラム(microSDログ保存あり版)
// バージョン   Ver.1.01
// Date         2023.11.06
// Copyright    ジャパンマイコンカーラリー実行委員会
// ライセンス   This software is released under the MIT License.
//              http://opensource.org/licenses/mit-license.php
//**************************************************************************

//======================================
// インクルード
//======================================
#include "mcr_gpt_lib.h"
#include <SPI.h>
#include <SD.h>

//======================================
// シンボル定義
//======================================
const char *C_DATE = __DATE__;          // microSDのファイル日付用 コンパイルした日付
const char *C_TIME = __TIME__;          //                         コンパイルした時間

// モータ、サーボのPWM関係
#define PWM_CYCLE     47999             //  16ms/(1/HOCO*16) = 0.016 /(1/48e6*16) = 48000
#define SERVO_CENTER  4499              // 1.5ms/(1/HOCO*16) = 0.0015/(1/48e6*16) =  4500
#define HANDLE_STEP   26                // １度あたりの値

// マスク値設定 ×：マスクあり(無効)　○：マスク無し(有効)
#define MASK2_2     0x66                // ×○○××○○×
#define MASK2_0     0x60                // ×○○×××××
#define MASK0_2     0x06                // ×××××○○×
#define MASK3_3     0xe7                // ○○○××○○○
#define MASK0_3     0x07                // ×××××○○○
#define MASK3_0     0xe0                // ○○○×××××
#define MASK4_0     0xf0                // ○○○○××××
#define MASK0_4     0x0f                // ××××○○○○
#define MASK4_4     0xff                // ○○○○○○○○

// ピン名定義
#define LED_D3      (D24)
#define DIP_SW1     (D25)
#define DIP_SW2     (D26)

// ログ処理関係
#define LOG_MAX     (1000)               // LOG_MAX × 10ms分のログをRAMに貯める ログが飛ぶときは増やしてください
#define LOG_NON     (0)
#define LOG_INIT    (1)
#define LOG_WRITE   (2)
#define LOG_END     (3)

//======================================
// プロトタイプ宣言
//======================================
void AGTCallback( timer_callback_args_t __attribute((unused)) * p_args );
void timer( unsigned long timer_set );
unsigned char sensor_inp( unsigned char mask );
int check_crossline( void );
int check_rightline( void );
int check_leftline( void );
unsigned char dipsw_get( void );
unsigned char pushsw_get( void );
unsigned char startbar_get( void );
void led_out( unsigned char led );
void motor( int accele_l, int accele_r );
void handle( int angle );
unsigned long convertBCD_CharToLong( unsigned char hex );

//======================================
// グローバル変数の宣言
//======================================
// 共通
volatile int f_setup_end;
volatile int pattern;
volatile unsigned long cnt0;
volatile unsigned long cnt1;
volatile unsigned long cnt_start;
volatile unsigned long cnt_print;

// 割り込み関係
FspTimer fsp_timer;

// microSD、ログ関係
File microSD;
volatile int err_sd;
volatile int log_st;
volatile int log_write;
volatile int log_read;
typedef struct {
    unsigned int time;  // 時間
    int ptn;            // パターン
    unsigned char sen;  // センサー
    char ang;           // 角度
    char ml;            // モータ左
    char mr;            // モータ右
} log_t;
volatile log_t log_buff[LOG_MAX];
volatile log_t log_raw;

//**********************************************************************
// setup
//**********************************************************************
void setup()
{
  setGPTterminal( 4, 6 );           // P406 GTIOC1B :サーボPWMとして使用
  setGPTterminal( 4, 5 );           // P405 GTIOC1A :左モータPWMとして使用
  setGPTterminal( 4, 3 );           // P403 GTIOC3A :右モータPWMとして使用

  GTIOC1A = 0;
  GTIOC1B = SERVO_CENTER;
  startPWM_GPT1( GTIOCA | GTIOCB , DIV16, PWM_CYCLE );  // GPT1 GTIOC1Aと1Bを使用 分周:16 周期:PWM_CYCLE

  GTIOC3A = 0;
  startPWM_GPT3( GTIOCA          , DIV16, PWM_CYCLE );  // GPT3 GTIOC3Aを使用     分周:16 周期:PWM_CYCLE

  // RMC-RA4M1基板上のLEDとDIPスイッチ
  pinMode( LED_D3, OUTPUT );
  pinMode( DIP_SW1, INPUT );
  pinMode( DIP_SW2, INPUT );

  // CN5 モータドライブ基板
  //pinMode( 27 , OUTPUT );         // LED2 rev.1.0bは入力端子なのでLEDは点灯しない
  //pinMode( 28 , OUTPUT );         // LED3 rev.1.0bは入力端子なのでLEDは点灯しない
  //pinMode( 29 , OUTPUT );         // サーボPWMなので設定しない
  //pinMode( 30 , OUTPUT );         // 右モータPWMなので設定しない
  pinMode( 31, OUTPUT );            // 右モータ方向
  //pinMode( 32 , OUTPUT );         // 左モータPWMなので設定しない
  pinMode( 33, OUTPUT );            // 左モータ方向
  pinMode( 34, INPUT_PULLUP);       // プッシュスイッチ

  // CN7 センサ基板
  pinMode( 35, INPUT );
  pinMode( 36, INPUT );
  pinMode( 37, INPUT );
  pinMode( 38, INPUT );
  pinMode( 39, INPUT );
  pinMode( 40, INPUT );
  pinMode( 41, INPUT );
  pinMode( 42, INPUT );

  // AGT 1msごとの割り込み処理の設定 PCLKB=24MHz ∴TIMER_SOURCE_DIV_1(1分周)なら、1/(24e6*1) * 24000 = 1ms  設定は１小さい値である23999を設定する
  fsp_timer.begin( TIMER_MODE_PERIODIC, AGT_TIMER, 1, 23999, 1, (timer_source_div_t)TIMER_SOURCE_DIV_1, AGTCallback );
  IRQManager::getInstance().addPeripheral( IRQ_AGT, (void*)fsp_timer.get_cfg() );
  fsp_timer.open();

  handle( 0 );              // ハンドルとモータの初期状態セット
  motor( 0, 0 );
  
  Serial.begin( 9600 );     // シリアル通信 速度9600bps

#if 0
  // シリアルが接続されるまで３秒待つ、３秒経っても接続されない場合はタイムアウト
  int i = 0;
  while ( !Serial ) {
    timer( 100 );
    i++;
    if( i >= 30 ) {
      break;
    }
  }
#endif

  R_PORT3->PODR_b.PODR5  = 1; // LED_D3 消灯

  Serial.print( "microSDを確認中です... " );

  if( !SD.begin(CS) ) {       // SDカードのチップセレクトライン それ以外はpins_arduino.hで定義
    err_sd = 1;
    Serial.println( "microSDが認識されません!" );
  } else {
    // microSD認識OK!
    SdFile::dateTimeCallback( &set_microSD_dateTime );  // コールバック関数の定義（日付と時刻を返す関数を登録）
    Serial.println( "microSDを認識しました。" );
  }

  // setup完了
  f_setup_end = 1;
}

//**********************************************************************
// loop：シリアル出力（デバッグ用）、microSDへのログ保存処理など、走行とは直接関係ない優先順位が低い処理
//**********************************************************************
void loop()
{
  static int log_pattern = 0;
  static int loopcnt = 0;
  char buf[16];
  int i;

#if 1
  // センサの状態をシリアル出力
  if( pattern == 0 && cnt_print >= 500 ) {
    cnt_print = 0;
    loopcnt++;
    unsigned char b = sensor_inp(MASK4_4);
    Serial.print( loopcnt ); // センサの状態の確認 シリアル出力
    Serial.print( " : sensor = " );
    if( (b & 0xf0) == 0x00 ) Serial.print( "0" );
    Serial.print( b, HEX );
    Serial.println( "" );
  }
#endif

  // microSDへのログ保存処理
  if( err_sd == 0 ) {
    switch( log_pattern ) {
      case 0:
        // 待機
        if( log_st == LOG_INIT ) {
          log_pattern = 1;
        }
        break;

      case 1:
        // 連番の読み込み
        i = 0;
        microSD = SD.open( "renban.txt", FILE_READ );
        if( microSD != 0 ) {
          int length = microSD.available();
          if( length > 8 ){
            length = 8;
          }
          microSD.read( buf, length );
          sscanf( buf, "%d", &i );
          if( i < 0 || i >= 9999 ) {
              i = 0;
          }
          microSD.close();
        } else {
          Serial.print( "renban.txtが開けませんので「log0001.csv」でファイルを作成します。" );
        }

        // 連番ファイルを一度、削除
        if( SD.exists( "renban.txt" ) ) {
          SD.remove( "renban.txt" );
        }

        // 連番ファイルを新規に作成
        microSD = SD.open( "renban.txt", FILE_WRITE );
        if( microSD != 0 ) {
          sprintf( buf, "%d", i + 1 );
          microSD.println( buf );
          microSD.close();
        }

        // 新しい番号でログファイル作成
        sprintf( buf, "log%04d.csv", i + 1 );
        microSD = SD.open( buf, FILE_WRITE );
        if( microSD != 0 ) {
          Serial.print( buf );
          Serial.println( "でログファイルを作成します。" );
          log_pattern = 2;
        } else {
          err_sd = 1;
          Serial.println( "ログファイルが作成できません。" );
        }
        break;

      case 2:
        // 最初に書き込む内容
        microSD.println( "ms,pattern,sensor,handle,motor_l,motor_r" );
        log_pattern = 3;
        break;

      case 3:
        if( log_st == LOG_WRITE ) {
          log_pattern = 11;
        }
        break;

      case 11:
        if( log_write != log_read ) {    // 保存するログが更新されたらmicroSDにログを書き込み
          sprintf( buf, "%08ld", convertBCD_CharToLong( log_buff[log_read].sen) );  // センサの値は2進数8桁で表示
          microSD.print( log_buff[log_read].time );
          microSD.print( "," );
          microSD.print( log_buff[log_read].ptn );
          microSD.print( ",'" );
          microSD.print( buf );
          microSD.print( "," );
          microSD.print( (int)log_buff[log_read].ang );
          microSD.print( "," );
          microSD.print( (int)log_buff[log_read].ml );
          microSD.print( "," );
          microSD.print( (int)log_buff[log_read].mr );
          microSD.println( "" );

          log_read++;
          if( log_read >= LOG_MAX ) {
            log_read = 0;
          }
        } else {
          if( log_st != LOG_WRITE ) {   // 最新のログが保存されていて書き込みでないなら終了
            // ログ処理終了
            log_pattern = 101;
          }
        }
        break;

      case 101:
        microSD.close();  // 必ずcloseしないと、ファイルが残らない
        log_pattern = 102;
        break;

      case 102:
        // 何もしない
        break;
    }
  }
}

//******************************************************************
// 1msごとの割り込み（通常走行、1msごとに実行する処理）
//******************************************************************
void AGTCallback( timer_callback_args_t __attribute((unused)) * p_args )
{
  int cnt10;
  unsigned char b;

  cnt0++;
  cnt1++;
  cnt_start++;
  cnt_print++;

  if( f_setup_end == 0 ) {  // セットアップ実行中はここで終わり
    return;
  }

  cnt10 = cnt_start % 10;
  if( cnt10 == 0 ) {     // 10msごとにログ保存
    if( log_st == LOG_WRITE) {
      log_buff[log_write].time = cnt_start;
      log_buff[log_write].ptn  = pattern;
      log_buff[log_write].sen  = sensor_inp(MASK4_4);
      log_buff[log_write].ang  = log_raw.ang;
      log_buff[log_write].ml   = log_raw.ml;
      log_buff[log_write].mr   = log_raw.mr;
      log_write++;
      if( log_write >= LOG_MAX ) {
        log_write = 0;
      }
    }
  }

  // 走行中にプッシュスイッチを押したら、走行終了
  if( pattern >= 11 && pattern <=1000 && cnt_start >= 1000 && pushsw_get() == 1 ) {
      pattern = 1001;
  }

  switch( pattern ) {

    //****************************************************************
    // パターンについて
    //  0：スイッチ入力待ち
    //  1：スタートバーが開いたかチェック
    // 11：通常トレース
    // 12：右へ大曲げの終わりのチェック
    // 13：左へ大曲げの終わりのチェック
    // 21：クロスライン検出時の処理
    // 22：クロスラインを読み飛ばす
    // 23：クロスライン後のトレース、クランク検出
    // 31：左クランククリア処理　安定するまで少し待つ
    // 32：左クランククリア処理　曲げ終わりのチェック
    // 33：左クランククリア処理　外側の白線と見間違わないようにする
    // 41：右クランククリア処理　安定するまで少し待つ
    // 42：右クランククリア処理　曲げ終わりのチェック
    // 43：右クランククリア処理　外側の白線と見間違わないようにする
    // 51：右ハーフライン検出時の処理
    // 52：右ハーフラインを読み飛ばす
    // 53：右ハーフライン後のトレース、レーンチェンジ
    // 54：右レーンチェンジ終了のチェック
    // 61：左ハーフライン検出時の処理
    // 62：左ハーフラインを読み飛ばす
    // 63：左ハーフライン後のトレース、レーンチェンジ
    // 64：左レーンチェンジ終了のチェック
    //****************************************************************

    case 0:
      // スイッチ入力待ち
      if( pushsw_get() == 1 ) {
        log_st = LOG_INIT;  // ログ処理開始（ファイルオープンなど)
        pattern = 1;
        cnt1 = 0;
        break;
      }

      // LED点滅周期の設定
      if( err_sd == 0 ) {   // microSDエラーなし
        if( cnt1 >= 500 ) {
          cnt1 = 0;
        }
        if( cnt1 < 50 ) {
          led_out( 0x1 ); // led_out( 0x1 );
        } else {
          led_out( 0x0 ); // led_out( 0x2 );
        }
      } else {
        if( cnt1 >= 200 ) { // LED点滅処理
          cnt1 = 0;
        }
        if( cnt1 < 100 ) {  // microSDエラーありは早く点滅
          led_out( 0x1 );  // led_out( 0x1 );
        } else {
          led_out( 0x0 );  // led_out( 0x2 );
        }
      }
      break;

    case 1:
      // スタートバーが開いたかチェック
      if( startbar_get() == 0 ) {
        // スタート！！
        led_out( 0x0 );
        cnt_start = 0;
        log_st = LOG_WRITE; // ログ保存開始
        pattern = 11;
        cnt1 = 0;
        break;
      }
      if( cnt1 >= 100 ) { // LED点滅処理
        cnt1 = 0;
      }
      if( cnt1 < 50 ) {
        led_out( 0x1 );  // led_out( 0x1 );
      } else {
        led_out( 0x0 );  // led_out( 0x2 );
      }
      break;

    case 11:
      // 通常トレース
      if( check_crossline() == 1 ) { // クロスラインチェック
        pattern = 21;
        break;
      }
      if( check_rightline() == 1 ) { // 右ハーフラインチェック
        pattern = 51;
        break;
      }
      if( check_leftline() == 1 ) { // 左ハーフラインチェック
        pattern = 61;
        break;
      }
      switch( sensor_inp(MASK3_3) ) {
        case 0x00:
          // センタ→まっすぐ
          handle( 0 );
          motor( 100, 100 );
          break;

        case 0x04:
          // 微妙に左寄り→右へ微曲げ
          handle( 5 );
          motor( 100, 100 );
          break;

        case 0x06:
          // 少し左寄り→右へ小曲げ
          handle( 10 );
          motor( 80, 69 );
          break;

        case 0x07:
          // 中くらい左寄り→右へ中曲げ
          handle( 15 );
          motor( 50, 38 );
          break;

        case 0x03:
          // 大きく左寄り→右へ大曲げ
          handle( 25 );
          motor( 30, 19 );
          pattern = 12;
          break;

        case 0x20:
          // 微妙に右寄り→左へ微曲げ
          handle( -5 );
          motor( 100, 100 );
          break;

        case 0x60:
          // 少し右寄り→左へ小曲げ
          handle( -10 );
          motor( 69, 80 );
          break;

        case 0xe0:
          // 中くらい右寄り→左へ中曲げ
          handle( -15 );
          motor( 38, 50 );
          break;

        case 0xc0:
          // 大きく右寄り→左へ大曲げ
          handle( -25 );
          motor( 19, 30 );
          pattern = 13;
          break;

        default:
          break;
      }
      break;

    case 12:
      // 右へ大曲げの終わりのチェック
      if( check_crossline() == 1 ) { // 大曲げ中もクロスラインチェック
        pattern = 21;
        break;
      }
      if( check_rightline() == 1 ) { // 右ハーフラインチェック
        pattern = 51;
        break;
      }
      if( check_leftline() == 1 ) { // 左ハーフラインチェック
        pattern = 61;
        break;
      }
      if( sensor_inp(MASK3_3) == 0x06 ) {
        pattern = 11;
      }
      break;

    case 13:
      // 左へ大曲げの終わりのチェック
      if( check_crossline() == 1 ) { // 大曲げ中もクロスラインチェック
        pattern = 21;
        break;
      }
      if( check_rightline() == 1 ) { // 右ハーフラインチェック
        pattern = 51;
        break;
      }
      if( check_leftline() == 1 ) { // 左ハーフラインチェック
        pattern = 61;
        break;
      }
      if( sensor_inp(MASK3_3) == 0x60 ) {
        pattern = 11;
      }
      break;

    case 21:
      // クロスライン検出時の処理
      led_out( 0x1 );  // led_out( 0x3 );
      handle( 0 );
      motor( 0, 0 );
      pattern = 22;
      cnt1 = 0;
      break;

    case 22:
      // クロスラインを読み飛ばす
      if( cnt1 >= 50 ) {
        pattern = 23;
        cnt1 = 0;
      }
      break;

    case 23:
      // クロスライン後のトレース、クランク検出
      if( sensor_inp(MASK4_0) == 0xf0 ) {
        // 左クランクと判断→左クランククリア処理へ
        led_out( 0x1 );  // led_out( 0x1 );
        handle( -38 );
        motor( 0, 50 );
        pattern = 31;
        cnt1 = 0;
        break;
      }
      if( sensor_inp(MASK0_4) == 0x0f ) {
        // 右クランクと判断→右クランククリア処理へ
        led_out( 0x1 );  // led_out( 0x2 );
        handle( 38 );
        motor( 50, 0 );
        pattern = 41;
        cnt1 = 0;
        break;
      }
      switch( sensor_inp(MASK3_3) ) {
        case 0x00:
          // センタ→まっすぐ
          handle( 0 );
          motor( 40, 40 );
          break;
        case 0x04:
        case 0x06:
        case 0x07:
        case 0x03:
          // 左寄り→右曲げ
          handle( 8 );
          motor( 40, 35 );
          break;
        case 0x20:
        case 0x60:
        case 0xe0:
        case 0xc0:
          // 右寄り→左曲げ
          handle( -8 );
          motor( 35, 40 );
          break;
      }
      break;

    case 31:
      // 左クランククリア処理　安定するまで少し待つ
      if( cnt1 >= 50 ) {
        pattern = 32;
        cnt1 = 0;
      }
      break;

    case 32:
      // 左クランククリア処理　曲げ終わりのチェック
      if( sensor_inp(MASK3_3) == 0x60 ) {
        led_out( 0x0 );
        pattern = 11;
        cnt1 = 0;
      }
      if( sensor_inp(MASK3_3) == 0x07 ) {  // 外側の白線を検出
        pattern = 33;
      }
      break;

    case 33:
      // 左クランククリア処理　外側の白線と見間違わないようにする
      b = sensor_inp(MASK3_3);
      if( b == 0x83 || b == 0x81 || b == 0xc1 ) {
        pattern = 32;
      }
      break;

    case 41:
      // 右クランククリア処理　安定するまで少し待つ
      if( cnt1 >= 50 ) {
        pattern = 42;
        cnt1 = 0;
      }
      break;

    case 42:
      // 右クランククリア処理　曲げ終わりのチェック
      if( sensor_inp(MASK3_3) == 0x06 ) {
        led_out( 0x0 );
        pattern = 11;
        cnt1 = 0;
      }
      if( sensor_inp(MASK3_3) == 0xe0 ) {  // 外側の白線を検出
        pattern = 43;
      }
      break;

    case 43:
      // 右クランククリア処理　外側の白線と見間違わないようにする
      b = sensor_inp(MASK3_3);
      if( b == 0xc1 || b == 0x81 || b == 0x83 ) {
        pattern = 42;
      }
      break;

    case 51:
      // 右ハーフライン検出時の処理
      led_out( 0x1 );  // led_out( 0x2 );
      handle( 0 );
      motor( 0, 0 );
      pattern = 52;
      cnt1 = 0;
      break;

    case 52:
      // 右ハーフラインを読み飛ばす
      if( cnt1 >= 50 ) {
        pattern = 53;
        cnt1 = 0;
      }
      if( check_crossline() == 1 ) {
        pattern = 21;
      }
      break;

    case 53:
      // 右ハーフライン後のトレース、レーンチェンジ
      if( sensor_inp(MASK4_4) == 0x00 ) {
        handle( 20 );
        motor( 40, 20 );
        pattern = 54;
        cnt1 = 0;
        break;
      }
      switch( sensor_inp(MASK3_3) ) {
        case 0x00:
          // センタ→まっすぐ
          handle( 0 );
          motor( 40, 40 );
          break;
        case 0x04:
        case 0x06:
        case 0x07:
        case 0x03:
          // 左寄り→右曲げ
          handle( 8 );
          motor( 40, 35 );
          break;
        case 0x20:
        case 0x60:
        case 0xe0:
        case 0xc0:
          // 右寄り→左曲げ
          handle( -8 );
          motor( 35, 40 );
          break;
        default:
          break;
      }
      break;

    case 54:
      // 右レーンチェンジ終了のチェック
      b = sensor_inp(MASK4_4);
      if( b == 0x3c || b == 0x1c || b == 0x38 ) {
        led_out( 0x0 );
        pattern = 11;
        cnt1 = 0;
      }
      break;

    case 61:
      // 左ハーフライン検出時の処理
      led_out( 0x1 );
      handle( 0 );
      motor( 0, 0 );
      pattern = 62;
      cnt1 = 0;
      break;

    case 62:
      // 左ハーフラインを読み飛ばす
      if( cnt1 >= 50 ) {
        pattern = 63;
        cnt1 = 0;
      }
      if( check_crossline() == 1 ) {
        pattern = 21;
      }
      break;

    case 63:
      // 左ハーフライン後のトレース、レーンチェンジ
      if( sensor_inp(MASK4_4) == 0x00 ) {
        handle( -20 );
        motor( 20, 40 );
        pattern = 64;
        cnt1 = 0;
        break;
      }
      switch( sensor_inp(MASK3_3) ) {
        case 0x00:
          // センタ→まっすぐ
          handle( 0 );
          motor( 40, 40 );
          break;
        case 0x04:
        case 0x06:
        case 0x07:
        case 0x03:
          // 左寄り→右曲げ
          handle( 8 );
          motor( 40, 35 );
          break;
        case 0x20:
        case 0x60:
        case 0xe0:
        case 0xc0:
          // 右寄り→左曲げ
          handle( -8 );
          motor( 35, 40 );
          break;
        default:
          break;
      }
      break;

    case 64:
      // 左レーンチェンジ終了のチェック
      b = sensor_inp(MASK4_4);
      if( b == 0x38 || b == 0x1c || b == 0x3c ) {
        led_out( 0x0 );
        pattern = 11;
        cnt1 = 0;
      }
      break;

    case 1001:
      // 走行終了処理
      log_st = LOG_END;

      motor( 0, 0 );
      if( cnt1 >= 1000 ) { // LED点滅処理
        cnt1 = 0;
      }
      if( cnt1 < 500 ) {
        led_out( 0x1 );  // led_out( 0x1 );
      } else {
        led_out( 0x0 );  // led_out( 0x2 );
      }
      break;
  }
}

//******************************************************************
// タイマ
// 引数　 タイマの時間 [ms]
//******************************************************************
void timer(unsigned long timer_set)
{
  cnt0 = 0;
  while( cnt0 < timer_set ) {
    // 時間になるまで待つ
  }
}

//******************************************************************
// センサ状態検出
// 引数　 マスク値
// 戻り値 センサ値
//******************************************************************
unsigned char sensor_inp( unsigned char mask )
{
  unsigned char sensor;

  sensor  = (!R_PORT0->PIDR_b.PIDR11 << 7);  // D35
  sensor |= (!R_PORT0->PIDR_b.PIDR10 << 6);  // D36
  sensor |= (!R_PORT0->PIDR_b.PIDR8  << 5);  // D37
  sensor |= (!R_PORT0->PIDR_b.PIDR7  << 4);  // D38
  sensor |= (!R_PORT0->PIDR_b.PIDR6  << 3);  // D39
  sensor |= (!R_PORT0->PIDR_b.PIDR5  << 2);  // D40
  sensor |= (!R_PORT0->PIDR_b.PIDR4  << 1);  // D41
  sensor |= (!R_PORT0->PIDR_b.PIDR3  << 0);  // D42

  sensor &= mask;

  return sensor;
}

//******************************************************************
// クロスライン検出処理
// 戻り値 0:クロスラインなし 1:あり
//******************************************************************
int check_crossline( void )
{
  unsigned char b;
  int ret;

  ret = 0;
  b = sensor_inp(MASK3_3);
  if( b == 0xe7 ) {
    ret = 1;
  }
  return ret;
}

//******************************************************************
// 右ハーフライン検出処理
// 戻り値 0:なし 1:あり
//******************************************************************
int check_rightline( void )
{
  unsigned char b;
  int ret = 0;

  b = sensor_inp(MASK4_4);
  if( b == 0x0f || b == 0x1f || b == 0x3f || b == 0x7f ) {
    ret = 1;
  }
  return ret;
}

//******************************************************************
// 左ハーフライン検出処理
// 戻り値 0:なし 1:あり
//******************************************************************
int check_leftline( void )
{
  unsigned char b;
  int ret = 0;

  b = sensor_inp(MASK4_4);
  if( b == 0xf0 || b == 0xf8 || b == 0xfc || b == 0xfe ) {
    ret = 1;
  }
  return ret;
}

//******************************************************************
// ディップスイッチ値読み込み
// 戻り値 スイッチ値 0～3
//******************************************************************
unsigned char dipsw_get( void )
{
  unsigned char sw;

  sw = ( !(R_PORT3->PIDR_b.PIDR7) << 1 ) + ( !(R_PORT3->PIDR_b.PIDR6) );

  return sw;
}

//*******************************************************************
// プッシュスイッチ値読み込み
// 戻り値 スイッチ値 ON:1 OFF:0
//*******************************************************************
unsigned char pushsw_get( void )
{
  unsigned char sw;

  sw = !(R_PORT4->PIDR_b.PIDR1);

  return sw;
}

//*******************************************************************
// スタートバー検出センサ読み込み
// 戻り値 センサ値 ON(バーあり):1 OFF(なし):0
//*******************************************************************
unsigned char startbar_get( void )
{
  unsigned char b;

  b = !(R_PORT0->PIDR_b.PIDR3);

  return b;
}

//*******************************************************************
// LED制御
// 引数　スイッチ値 LED3 "0":消灯 "1":点灯
//*******************************************************************
void led_out( unsigned char led )
{
  R_PORT3->PODR_b.PODR5 = !led;
}

//*******************************************************************
// モータ速度制御
// 引数　 左モータ:-100～100、右モータ:-100～100
//        0で停止、100で正転100%、-100で逆転100%
// 戻り値 なし
//*******************************************************************
void motor( int accele_l, int accele_r )
{
  int sw_data;

  sw_data  = dipsw_get() + 7;  // 7～10
  accele_l = accele_l * sw_data / 10;
  accele_r = accele_r * sw_data / 10;

  log_raw.ml = accele_l;
  log_raw.mr = accele_r;

  // 左モータ制御
  if( accele_l >= 0 ) {
    R_PORT4->PODR_b.PODR2 = 0;
    accele_l = (long)(PWM_CYCLE + 1) * accele_l / 100;
    GTIOC3A = accele_l;
  } else {
    R_PORT4->PODR_b.PODR2 = 1;
    accele_l = (long)(PWM_CYCLE + 1) * (-accele_l) / 100;
    GTIOC3A = accele_l;
  }

  // 右モータ制御
  if( accele_r >= 0 ) {
    R_PORT4->PODR_b.PODR4 = 0;
    accele_r = (long)(PWM_CYCLE + 1) * accele_r / 100;
    GTIOC1A = accele_r;
  } else {
    R_PORT4->PODR_b.PODR4 = 1;
    accele_r = (long)(PWM_CYCLE + 1) * (-accele_r) / 100;
    GTIOC1A = accele_r;
  }
}

//*******************************************************************
// サーボハンドル操作
// 引数　 サーボ操作角度：-90～90
//        -90で左へ90度、0でまっすぐ、90で右へ90度回転
//*******************************************************************
void handle( int angle )
{
  log_raw.ang = angle;
  // サーボが左右逆に動く場合は、「-」を「+」に替えてください
  GTIOC1B = SERVO_CENTER - angle * HANDLE_STEP;
}

//**********************************************************************
// microSD.close時に、ファイルの日付・時刻を設定
//**********************************************************************
void set_microSD_dateTime( uint16_t* date, uint16_t* time )
{
  // FAT_DATEマクロでフィールドを埋めて日付を返す
  *date = FAT_DATE( (uint16_t)getCompileYear( C_DATE ),
                    (uint8_t)getCompileMonth( C_DATE ),
                    (uint8_t)getCompileDay( C_DATE ) );

  // FAT_TIMEマクロでフィールドを埋めて時間を返す
  *time = FAT_TIME( getCompileHour( C_TIME ),
                    getCompilerMinute( C_TIME ),
                    getCompilerSecond( C_TIME ) );
}

//**********************************************************************
// char型データの値をlong型変数に2進数で変換
// 引数　 unsigned char 変換元の8bitデータ
// 戻り値 unsigned long 変換先の変数(0～11111111) ※0か1しかありません
//**********************************************************************
unsigned long convertBCD_CharToLong( unsigned char hex )
{
  int             i;
  unsigned long   l = 0;

  for( i=0; i<8; i++ ) {
    l *= 10;
    if( (hex & 0x80) != 0 ) {
      l += 1;
    }
    hex <<= 1;
  }

  return l;
}

//**********************************************************************
// モジュール名 getCompileYear
// 処理概要     コンパイルした時の年を取得
// 引数　       なし
// 戻り値       年
//**********************************************************************
int getCompileYear( const char *p )
{
  int i;

  i = atoi( p + 7 );
  if( i < 1980 || i > 2107 ) {
    i = 2023;
  }

  return i;
}

//**********************************************************************
// モジュール名 getCompileMonth
// 処理概要     コンパイルした時の月を取得
// 引数　       なし
// 戻り値       月
//**********************************************************************
int getCompileMonth( const char *p )
{
  const char monthStr[] = { "JanFebMarAprMayJunJulAugSepOctNovDec" };
  int i, r;

  for( i=0; i<12; i++ ) {
    r = strncmp( monthStr + i * 3, p, 3 );
    if( r == 0 ) {
      return i + 1;
    }
  }

  return 1;
}

//**********************************************************************
// モジュール名 getCompileDay
// 処理概要     コンパイルした時の日を取得
// 引数　       なし
// 戻り値       日
//**********************************************************************
int getCompileDay( const char *p )
{
  int i;

  i = atoi( p + 4 );
  if( i < 1 || i > 31 ) {
    i = 1;
  }

  return i;
}

//**********************************************************************
// モジュール名 getCompileHour
// 処理概要     コンパイルした時の時を取得
// 引数　       なし
// 戻り値       時
//**********************************************************************
int getCompileHour( const char *p )
{
  int i;

  i = atoi( p );

  if( i < 0 || i > 23 ) {
    i = 0;
  }

  return i;
}

//**********************************************************************
// モジュール名 getCompilerMinute
// 処理概要     コンパイルした時の分を取得
// 引数　       なし
// 戻り値       分
//**********************************************************************
int getCompilerMinute( const char *p )
{
  int i;

  i = atoi( p + 3 );
  if( i < 0 || i > 59 ) {
    i = 0;
  }

  return i;
}

//**********************************************************************
// モジュール名 getCompilerSecond
// 処理概要     コンパイルした時の秒を取得
// 引数　       なし
// 戻り値       秒
//**********************************************************************
int getCompilerSecond( const char *p )
{
  int i;

  i = atoi( p + 6 );

  if( i < 0 || i > 59 ) {
    i = 0;
  }

  return i;
}

//**********************************************************************
// end of file
//**********************************************************************

/*
改訂内容
Ver.1.00 2023.09.04 作成
Ver.1.01 2023.11.06 ハンドル角度、モータ値をkit12_38a.cと同じにする
*/
